//	Altirra - Atari 800/800XL/5200 emulator
//	Copyright (C) 2024 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License along
//	with this program. If not, see <http://www.gnu.org/licenses/>.

#include <stdafx.h>
#include <at/atcore/audiomixer.h>
#include "printerbase.h"

ATPrinterSoundSource::ATPrinterSoundSource() {
	const float volume = 0.2f;

	mpCurrentEdgeBuffer = new ATSyncAudioEdgeBuffer;
	mpCurrentEdgeBuffer->mLeftVolume = volume;
	mpCurrentEdgeBuffer->mRightVolume = volume;

	mpNextEdgeBuffer = new ATSyncAudioEdgeBuffer;
	mpNextEdgeBuffer->mLeftVolume = volume;
	mpNextEdgeBuffer->mRightVolume = volume;
}

ATPrinterSoundSource::~ATPrinterSoundSource() {
	VDASSERT(!mpAudioMixer);
}

void ATPrinterSoundSource::Init(IATAudioMixer& mixer, const char *debugName) {
	VDASSERT(!mpAudioMixer);

	mpCurrentEdgeBuffer->mpDebugLabel = debugName;
	mpNextEdgeBuffer->mpDebugLabel = debugName;

	mpAudioMixer = &mixer;
	mpAudioMixer->AddSyncAudioSource(this);
}

void ATPrinterSoundSource::Shutdown() {
	if (mpAudioMixer) {
		mpAudioMixer->RemoveSyncAudioSource(this);
		mpAudioMixer = nullptr;
	}
}

void ATPrinterSoundSource::AddPinSound(uint32 t, int numPins) {
	// don't keep readding events on the same cycle (only possible through debugger)
	if (mPinTimes.empty() || mPinTimes.back().mTime != t)
		mPinTimes.emplace_back(t, 1);
}

bool ATPrinterSoundSource::RequiresStereoMixingNow() const {
	return false;
}

void ATPrinterSoundSource::WriteAudio(const ATSyncAudioMixInfo& mixInfo) {
	const uint32 baseTime = (uint32)mixInfo.mStartTime;
	const uint32 cycleWindow = mixInfo.mNumCycles;

	auto& currentEdges = mpCurrentEdgeBuffer->mEdges;
	auto& nextEdges = mpNextEdgeBuffer->mEdges;

	for(const auto& pinEvent : mPinTimes) {
		const uint32 startT = pinEvent.mTime + 128;
		const uint32 stopT = startT + mPulseWidth;
		const uint32 relStartT = (uint32)(startT - baseTime);
		const uint32 relStopT = (uint32)(stopT - baseTime);
		const bool inWindowStart = relStartT < cycleWindow;
		const bool inWindowStop = relStopT < cycleWindow;

		VDASSERT(relStartT < cycleWindow * 2);

		auto& startEdge = (inWindowStart ? currentEdges : nextEdges).emplace_back();
		startEdge.mTime = startT;
		startEdge.mDeltaValue = (float)pinEvent.mNumPins;

		auto& stopEdge = (inWindowStop ? currentEdges : nextEdges).emplace_back();
		stopEdge.mTime = stopT;
		stopEdge.mDeltaValue = -(float)pinEvent.mNumPins;
	}

	mPinTimes.clear();

	mpAudioMixer->GetEdgePlayer().AddEdgeBuffer(mpCurrentEdgeBuffer);

	std::swap(mpCurrentEdgeBuffer, mpNextEdgeBuffer);
}
